@@ -1,5 +1,5 @@
 /* Dump a gcov file, for debugging use.
-   Copyright (C) 2002, 2003 Free Software Foundation, Inc.
+   Copyright (C) 2002, 2003, 2004, 2005, 2006 Free Software Foundation, Inc.
    Contributed by Nathan Sidwell <nathan@codesourcery.com>
 
 Gcov is free software; you can redistribute it and/or modify
@@ -77,6 +77,9 @@ main (int argc ATTRIBUTE_UNUSED, char **argv)
 {
   int opt;
 
+  /* Unlock the stdio streams.  */
+  unlock_std_streams ();
+
   while ((opt = getopt_long (argc, argv, "hlpv", options, NULL)) != -1)
     {
       switch (opt)
@@ -118,7 +121,7 @@ static void
 print_version (void)
 {
   printf ("gcov-dump (GCC) %s\n", version_string);
-  printf ("Copyright (C) 2003 Free Software Foundation, Inc.\n");
+  printf ("Copyright (C) 2006 Free Software Foundation, Inc.\n");
   printf ("This is free software; see the source for copying conditions.\n"
   	  "There is NO warranty; not even for MERCHANTABILITY or \n"
 	  "FITNESS FOR A PARTICULAR PURPOSE.\n\n");
@@ -254,8 +257,6 @@ dump_file (const char *filename)
 	  break;
 	}
     }
-  if (!gcov_is_eof ())
-    printf ("%s:early end of file\n", filename);
   gcov_close ();
 }
 
