@@ -1,5 +1,5 @@
 /* Tree inlining hooks and declarations.
-   Copyright 2001, 2003 Free Software Foundation, Inc.
+   Copyright 2001, 2003, 2004 Free Software Foundation, Inc.
    Contributed by Alexandre Oliva  <aoliva@redhat.com>
 
 This file is part of GCC.
@@ -26,11 +26,11 @@ Boston, MA 02111-1307, USA.  */
 
 void optimize_inline_calls (tree);
 bool tree_inlinable_function_p (tree);
-tree walk_tree (tree*, walk_tree_fn, void*, void*);
-tree walk_tree_without_duplicates (tree*, walk_tree_fn, void*);
-tree copy_tree_r (tree*, int*, void*);
-void clone_body (tree, tree, void*);
-void remap_save_expr (tree*, void*, tree, int*);
+tree copy_tree_r (tree *, int *, void *);
+void clone_body (tree, tree, void *);
+tree save_body (tree, tree *, tree *);
+int estimate_move_cost (tree type);
+int estimate_num_insns (tree expr);
 
 /* 0 if we should not perform inlining.
    1 if we should expand functions calls inline at the tree level.
